@@ -10,11 +10,14 @@ from sklearn.metrics.pairwise import cosine_similarity
 from scipy.spatial.distance import jensenshannon
 from scipy.stats import entropy
 import warnings
-# warnings.filterwarnings('ignore')
 ###############################################################################
 
 
 def stringtime(n):
+    """
+    :param n: integer, representing a number seconds
+    :return: string representing time, in format hh:mm:ss
+    """
     h = str(int(n / 3600))
     m = str(int((n % 3600) / 60))
     s = str(int((n % 3600) % 60))
@@ -25,61 +28,40 @@ def stringtime(n):
 
 
 def start(sep=True):
-    start = time.time()
+    """
+    This function prints and returns the present moment
+    :param sep: bool. If True, a separation line is printed on the shell
+    :return: the present moment
+    """
+    stt = time.time()
     now = time.strftime("%Y/%m/%d %H:%M:%S")
     if sep: print('#'*80)
     print('start:', now)
-    return start
+    return stt
 
 
-def end(start, sep=True):
-    end = time.time()
-    dur = end - start
-    str_dur = stringtime(end - start)
-    now = time.strftime("%Y/%m/%d %H:%M:%S")
+def end(stt, sep=True):
+    """
+    The function takes an hour in the past and prints the present moment and the time elapsed from the input moment
+    :param stt: a time.time() variable
+    :param sep: bool. If True, a separation line is printed on the shell
+    :return: the number of seconds between the input hour and the present moment
+    """
+    now = time.time()
+    dur = now - stt
+    str_dur = stringtime(now - stt)
+    str_now = time.strftime("%Y/%m/%d %H:%M:%S")
     if sep:
-        print('#'*80 + "\nend:", now, " - time elapsed:", str_dur + "\n" + '#'*80)
+        print('#'*80 + "\nend:", str_now, " - time elapsed:", str_dur + "\n" + '#'*80)
     else:
-        print("end:", now, " - time elapsed:", str_dur)
+        print("end:", str_now, " - time elapsed:", str_dur)
     return dur
 
-
-###############################################################################
-
-
-def get_lines(path, maxrow=-1):
-    """the function traverses a generator until maxrow or StopIteration.
-    the output is the another generator, having the wanted nr of lines"""
-    generator = open(path)
-    row_counter = 0
-    while maxrow != 0:
-        try:
-            line = next(generator)
-            maxrow -= 1
-            row_counter += 1
-            yield line
-        except StopIteration:
-            print(f"{'nr rows in generator:':.<40} {row_counter}")
-            maxrow = 0
-
-
-def read_file(filename, code='utf-8'):
-    with open(filename, 'r', encoding=code) as f_in:
-        out = f_in.read()
-        return out
-
-
-def printjson(data, stop=None):
-    for i, k in enumerate(data):
-        if i == stop: break
-        print(f"{k}: {json.dumps(data[k], indent=4, ensure_ascii=False)}") # ensure ascii false permette di stampare i caratteri utf-8 invece di \usblindo
-    return 1
-
     
-###############################################################################
-
-
-class bcolors:
+class BColor:
+    """
+    Class that defines colors for fancy results' printing
+    """
     reset     = '\033[0m'
     bold      = '\033[1m'
     underline = '\033[4m'
@@ -149,31 +131,40 @@ class bcolors:
     CBEIGEBG2  = '\33[106m'
     CWHITEBG2  = '\33[107m'
 
-    i2red = {0:  ([255/255, 204/255, 204/255]), # quasi bianco
+    # red nuances:
+    i2red = {0:  ([255/255, 204/255, 204/255]), # almost white
              1:  ([255/255, 153/255, 153/255]),
              2:  ([255/255, 102/255, 102/255]),
              3:  ([255/255, 51/255,  51/255]),
-             4:  ([255/255, 0/255,   0/255]), # rosso
+             4:  ([255/255, 0/255,   0/255]), # red
              5:  ([204/255, 0/255,   0/255]),
              6:  ([153/255, 0/255,   0/255]),
              7:  ([102/255, 0/255,   0/255]),
-             8:  ([51/255,  0/255,   0/255])} # quasi nero
-
-    def examples(self):
-        for i in range(0, 16):
-            for j in range(0, 16):
-                code = str(i * 16 + j)
-                sys.stdout.write(u"\u001b[38;5;" + code + "m " + code.ljust(4))
-        for i in range(0, 16):
-            for j in range(0, 16):
-                code = str(i * 16 + j)
-                sys.stdout.write(u"\u001b[48;5;" + code + "m " + code.ljust(4))
+             8:  ([51/255,  0/255,   0/255])} # almost black
 
 
 ###############################################################################
 
+
 class Bootstrap:
+    """
+    Class that computes performance metrics for hard and soft labels,
+    and the bootstrap sampling significance test.
+    I takes experiments' results and it prints performance and significance
+    levels, saving both inputs and outputs to different files
+    """
     def __init__(self, save_results=True, save_outcomes=True, dir_out=''):
+        """
+        :param save_results: type: bool; default: True. If True, the results
+        (performance and significance levels) are saved as tsv file(s)
+        :param save_outcomes: type: bool; default: True. If True, the inputs
+        (experiments' outcomes) are saved as json file
+        :param dir_out: type: str; default: ''. Dir where to save the files above
+        :attribute data:
+                defaultdict(lambda: {'exp_idxs': list(), 'preds': list(), 'targs': list(), 'idxs': list(), 'epochs': list(),
+                                     'h1': defaultdict(lambda: {'exp_idxs': list(), 'preds': list(), 'targs': list(), 'idxs': list(), 'epochs': list()})}
+                Used to store and save to json the class inputs.
+        """
         self.dirout = dir_out + '/' if (dir_out != '') and (not re.search('/$', dir_out)) else dir_out
         self.savetsv = save_results
         self.savejson = save_outcomes
@@ -189,6 +180,20 @@ class Bootstrap:
                                                                     'epochs':   list()})})
 
     def feed(self, h0, h1=None, exp_idx=None, preds=None, targs=None, idxs=None, epochs=()):
+        """
+        :param h0: type: str. It identifies the control condition.
+        :param h1: type: str. It identifies the experimental condition.
+        :param exp_idx: type: str. It identifies the experiments' index
+        :param preds:  type: list, numpy.ndarray or str. If str, interpreted as a path to a .npy, .csv,
+        .tsv or .txt file, containing experiments' outcomes (hard or soft labels)
+        :param targs: type: list, numpy.ndarray or str. If str, interpreted as a path to a .npy, .csv,
+        .tsv or .txt file, containing experiments' targets (hard or soft labels)
+        :param idxs:  type: list, numpy.ndarray or str. If str, interpreted as a path to a .npy, .csv,
+        .tsv or .txt file, containing experiments' data point indexes
+        :param epochs: type: int. Number of experiment's training epochs.
+        :return: 1. The function stores the inputs in a dictionary accessible from other functions,
+        and that can be saved as json file
+        """
         targs = self.input2ndarray(targs)
         preds = self.input2ndarray(preds)
         idxs = np.arange(len(preds)) if idxs is None else self.input2ndarray(idxs)
@@ -208,6 +213,11 @@ class Bootstrap:
         return 1
 
     def data2json(self, pathname):
+        """
+        I takes the path where to save the input data and saves them to json.
+        :param pathname: path where to save the json file
+        :return: 1
+        """
         for h0 in self.data:
             self.data[h0]['targs'] = [targ.tolist() for targ in self.data[h0]['targs']]
             self.data[h0]['preds'] = [pred.tolist() for pred in self.data[h0]['preds']]
@@ -220,6 +230,12 @@ class Bootstrap:
         return 1
 
     def loadjson(self, pathname):
+        """
+        It takes a path to a json file previously saved and loads it in a dictionary
+        accessible from other functions
+        :param pathname: path to a json file previously saved.
+        :return: 1
+        """
         with open(pathname) as f_in: jin = json.load(f_in)
         for h0 in jin:
             self.data[h0]['exp_idxs'] = jin[h0]['exp_idxs']
@@ -238,6 +254,12 @@ class Bootstrap:
 
     @staticmethod
     def input2ndarray(obj):
+        """
+        It takes an input object and converts it in a 2D numpy.ndarray
+        :param obj: list, numpy.ndarray or str. If str, treated as a path to a .npy, .csv, .tsv
+        or .txt file.
+        :return: 2D numpy.ndarray
+        """
         if isinstance(obj, np.ndarray):
             if len(obj.shape) == 1:
                 return obj.reshape(-1, 1)
@@ -263,6 +285,12 @@ class Bootstrap:
 
     @staticmethod
     def crossentropy_mean(p_targ, q_pred):
+        """
+        Given 2 lists of probability distributions, computes the mean cross-entropy
+        :param p_targ: list of target probability ditributions
+        :param q_pred: list of predicted probability ditributions
+        :return: mean cross-entropy
+        """
         epsilon = 1e-12
         p_targ = np.clip(p_targ, epsilon, 1. - epsilon) # se qualche valore deborda, quello che eccede lo riduco tra 0espiccioli e 1-0espiccioli
         q_pred = np.clip(q_pred, epsilon, 1. - epsilon)
@@ -270,6 +298,18 @@ class Bootstrap:
         return ce.mean()
 
     def metrics(self, targs, h0_preds, h1_preds, h0_name='h0', h1_name='h1', targetclass=None, verbose=False):
+        """
+        :param targs: type: numpy.ndarray. target hard or soft labels
+        :param h0_preds: type: numpy.ndarray. predicted hard or soft labels from a control experiment
+        :param h1_preds: type: numpy.ndarray. predicted hard or soft labels from a treatment experiment
+        :param h0_name: type: str. it identifies the control condition
+        :param h1_name: type: str. it identifies the treatment condition
+        :param targetclass: type: int, default: None. If given, performance and significance levels
+         are computed wrt the class whose index is given
+        :param verbose: type: bool; default: False. If true, it prints performance information.
+        :return: two pandas DataFrame: the first contain the overall performance,
+        the second the target class performance (empty if not requested)
+        """
         if targs.shape[1] == 1:
             rounding_value = 2
             h0_acc  = accuracy_score(targs, h0_preds)
@@ -282,16 +322,17 @@ class Bootstrap:
             h1_rec  = recall_score(targs, h1_preds, average='macro')
             # h0_conf_matrix = confusion_matrix(targs, h0_preds)
             # h1_conf_matrix = confusion_matrix(targs, h1_preds)
+            # print(h0_conf_matrix, "\n", h1_conf_matrix)
             diff_acc  = h1_acc  - h0_acc
             diff_f1   = h1_f1   - h0_f1
             diff_prec = h1_prec - h0_prec
             diff_rec  = h1_rec  - h0_rec
-            df_tot = pd.DataFrame({'f1':    [h0_f1,       h1_f1],       'd_f1':    ['',  diff_f1],       's_f1':    ['', ''],
-                                   'acc':   [h0_acc,      h1_acc],      'd_acc':   ['',  diff_acc],      's_acc':   ['', ''],
-                                   'prec':  [h0_prec,     h1_prec],     'd_prec':  ['',  diff_prec],     's_prec':  ['', ''],
-                                   'rec':   [h0_rec,      h1_rec],      'd_rec':   ['',  diff_rec],      's_rec':   ['', ''],
-                                   'mean_epochs': [None, None]
-                                 }, index=[h0_name, h1_name])
+            df_tot = pd.DataFrame({ 'f1':    [h0_f1,       h1_f1],       'd_f1':    ['',  diff_f1],       's_f1':    ['', ''],
+                                    'acc':   [h0_acc,      h1_acc],      'd_acc':   ['',  diff_acc],      's_acc':   ['', ''],
+                                    'prec':  [h0_prec,     h1_prec],     'd_prec':  ['',  diff_prec],     's_prec':  ['', ''],
+                                    'rec':   [h0_rec,      h1_rec],      'd_rec':   ['',  diff_rec],      's_rec':   ['', ''],
+                                    'mean_epochs': [None, None]
+                                    }, index=[h0_name, h1_name])
             df_tot = df_tot['mean_epochs f1 d_f1 s_f1 acc d_acc s_acc prec d_prec s_prec rec d_rec s_rec'.split()]
             if verbose:
                 h0_countpreds = Counter(h0_preds.flatten())
@@ -305,9 +346,9 @@ class Bootstrap:
                 print(f"{'h0 preds count:':<15} {h0_countpreds}")
                 print(f"{'h1 preds count:':<15} {h1_countpreds}")
                 print(f"{'F-measure':.<15} - h0: {h0_f1:<7.4f} - h1: {h1_f1:<7.4f} - diff: {diff_f1:.4f}")
-                print(f"{'accuracy':.<15} - h0: {h0_acc:<7.4f} - h1: {h1_acc:<7.4f} - diff: {diff_acc:.4f}")
                 print(f"{'precision':.<15} - h0: {h0_prec:<7.4f} - h1: {h1_prec:<7.4f} - diff: {diff_prec:.4f}")
                 print(f"{'recall':.<15} - h0: {h0_rec:<7.4f} - h1: {h1_rec:<7.4f} - diff: {diff_rec:.4f}")
+                print(f"{'accuracy':.<15} - h0: {h0_acc:<7.4f} - h1: {h1_acc:<7.4f} - diff: {diff_acc:.4f}")
 
             df_tgt = pd.DataFrame(index=[h0_name, h1_name])
             if targetclass is not None:
@@ -323,11 +364,11 @@ class Bootstrap:
                 diff_tgt_f1   = round(h1_tgt_f1   - h0_tgt_f1,   rounding_value)
                 diff_tgt_prec = round(h1_tgt_prec - h0_tgt_prec, rounding_value)
                 diff_tgt_rec  = round(h1_tgt_rec  - h0_tgt_rec,  rounding_value)
-                df_tgt = pd.DataFrame({'tf1':   [h0_tgt_f1,   h1_tgt_f1],   'd_tf1':   ['',  diff_tgt_f1],   's_tf1':   ['', ''],
-                                       'tprec': [h0_tgt_prec, h1_tgt_prec], 'd_tprec': ['',  diff_tgt_prec], 's_tprec': ['', ''],
-                                       'trec':  [h0_tgt_rec,  h1_tgt_rec],  'd_trec':  ['',  diff_tgt_rec],  's_trec':  ['', ''],
-                                       'mean_epochs': [None, None]
-                                      }, index=[h0_name, h1_name])
+                df_tgt = pd.DataFrame({ 'tf1':   [h0_tgt_f1,   h1_tgt_f1],   'd_tf1':   ['',  diff_tgt_f1],   's_tf1':   ['', ''],
+                                        'tprec': [h0_tgt_prec, h1_tgt_prec], 'd_tprec': ['',  diff_tgt_prec], 's_tprec': ['', ''],
+                                        'trec':  [h0_tgt_rec,  h1_tgt_rec],  'd_trec':  ['',  diff_tgt_rec],  's_trec':  ['', ''],
+                                        'mean_epochs': [None, None]
+                                        }, index=[h0_name, h1_name])
                 df_tgt = df_tgt['mean_epochs tf1 d_tf1 s_tf1 tprec d_tprec s_tprec trec d_trec s_trec'.split()]
                 if verbose:
                     print(f"{'targetclass F-measure':.<25} - h0: {h0_tgt_f1:<7.4f} - h1: {h1_tgt_f1:<7.4f} - diff: {diff_tgt_f1:.4f}")
@@ -356,12 +397,12 @@ class Bootstrap:
             diff_sim = h1_sim - h0_sim
             diff_cor = h1_cor - h0_cor
 
-            df_tot = pd.DataFrame({'jsd': [h0_jsd, h1_jsd], 'd_jsd': ['',  diff_jsd], 's_jsd': ['', ''],
-                                   'ce':  [h0_ce,  h1_ce],  'd_ce':  ['',  diff_ce],  's_ce':  ['', ''],
-                                   'sim': [h0_sim, h1_sim], 'd_sim': ['',  diff_sim], 's_sim': ['', ''],
-                                   'cor': [h0_cor, h1_cor], 'd_cor': ['',  diff_cor], 's_cor': ['', ''],
-                                   'mean_epochs': [None, None]
-                                  }, index=[h0_name, h1_name])
+            df_tot = pd.DataFrame({ 'jsd': [h0_jsd, h1_jsd], 'd_jsd': ['',  diff_jsd], 's_jsd': ['', ''],
+                                    'ce':  [h0_ce,  h1_ce],  'd_ce':  ['',  diff_ce],  's_ce':  ['', ''],
+                                    'sim': [h0_sim, h1_sim], 'd_sim': ['',  diff_sim], 's_sim': ['', ''],
+                                    'cor': [h0_cor, h1_cor], 'd_cor': ['',  diff_cor], 's_cor': ['', ''],
+                                    'mean_epochs': [None, None]
+                                    }, index=[h0_name, h1_name])
             df_tot = df_tot['mean_epochs jsd d_jsd s_jsd ce d_ce s_ce sim d_sim s_sim cor d_cor s_cor'.split()]
             if verbose:
                 print(f"h0: {h0_name} - h1: {h1_name}")
@@ -385,10 +426,10 @@ class Bootstrap:
                 diff_tgt_jsd = h1_tgt_jsd - h0_tgt_jsd
                 diff_tgt_ce  = h1_tgt_ce  - h0_tgt_ce
 
-                df_tgt = pd.DataFrame({'tjsd': [h0_tgt_jsd, h1_tgt_jsd], 'd_tjsd': ['',  diff_tgt_jsd], 's_tjsd': ['', ''],
-                                       'tce':  [h0_tgt_ce,  h1_tgt_ce],  'd_tce':  ['',  diff_tgt_ce],  's_tce':  ['', ''],
-                                       'mean_epochs': [None, None]
-                                  }, index=[h0_name, h1_name])
+                df_tgt = pd.DataFrame({ 'tjsd': [h0_tgt_jsd, h1_tgt_jsd], 'd_tjsd': ['',  diff_tgt_jsd], 's_tjsd': ['', ''],
+                                        'tce':  [h0_tgt_ce,  h1_tgt_ce],  'd_tce':  ['',  diff_tgt_ce],  's_tce':  ['', ''],
+                                        'mean_epochs': [None, None]
+                                        }, index=[h0_name, h1_name])
                 df_tgt = df_tgt['mean_epochs tjsd d_tjsd s_tjsd tce d_tce s_tce'.split()]
                 if verbose:
                     print(f"target {targetclass} {'Jensen-Shannon divergence:':<26} - h0: {h0_tgt_jsd:<7.4f} - h1: {h1_tgt_jsd:<7.4f} - diff: {diff_tgt_jsd:.4f}")
@@ -397,6 +438,26 @@ class Bootstrap:
             return df_tot, df_tgt
 
     def test(self, targs, h0_preds, h1_preds, h0_name='h0', h1_name='h1', n_loops=1000, sample_size=.1, targetclass=None, verbose=False):
+        """
+        :param targs: type: list, numpy.ndarray or str. If str, interpreted as a path to a .npy, .csv,
+        .tsv or .txt file, containing experiments' targets (hard or soft labels)
+        :param h0_preds: type: list, numpy.ndarray or str. If str, interpreted as a path to a .npy, .csv,
+        .tsv or .txt file, containing experiments' predictions (hard or soft labels), from the control
+        condition (h0)
+        :param h1_preds: type: list, numpy.ndarray or str. If str, interpreted as a path to a .npy, .csv,
+        .tsv or .txt file, containing experiments' predictions (hard or soft labels), from the
+        experimental condition (h1)
+        :param h0_name: type: str. it identifies the control condition
+        :param h1_name: type: str. it identifies the treatment condition
+        :param n_loops: type: int; default: 1000. number of iterations for bootstrap sampling.
+        :param sample_size: type: float; default: .1. Percent sample size wrt the test set, f
+        or bootstrap sampling.
+        :param targetclass: type: int; default: None. Index of a target class. If given, performance and
+        significance levels are computed for that class.
+        :param verbose: type: bool; default: False. If true, it prints performance information.
+        :return: two pandas DataFrame: the first contain the overall performance,
+        the second the target class performance (empty if not requested)
+        """
         assert .05 <= sample_size <= .5, 'sample_size must be between .05 and .5'
         targs    = self.input2ndarray(targs)
         h0_preds = self.input2ndarray(h0_preds)
@@ -413,6 +474,7 @@ class Bootstrap:
             twice_diff_f1   = 0
             twice_diff_prec = 0
             twice_diff_rec  = 0
+            diff_tgt_f1, diff_tgt_prec, diff_tgt_rec, twice_diff_tgt_f1, twice_diff_tgt_prec, twice_diff_tgt_rec = None, None, None, None, None, None
             if targetclass is not None:
                 diff_tgt_f1   = df_tgt.d_tf1[-1]
                 diff_tgt_prec = df_tgt.d_tprec[-1]
@@ -435,26 +497,26 @@ class Bootstrap:
                     if df_sample_tgt.d_tf1[-1]    > 2 * diff_tgt_f1:   twice_diff_tgt_f1   += 1
                     if df_sample_tgt.d_tprec[-1]  > 2 * diff_tgt_prec: twice_diff_tgt_prec += 1
                     if df_sample_tgt.d_trec[-1]   > 2 * diff_tgt_rec:  twice_diff_tgt_rec  += 1
-            col_sign_f1   = f"{bcolors.red}**{bcolors.reset}" if twice_diff_f1   / n_loops < 0.01 else f"{bcolors.red}*{bcolors.reset}" if twice_diff_f1   / n_loops < 0.05 else f"{bcolors.grey}!{bcolors.reset}" if twice_diff_f1   / n_loops > 0.95 else f"{bcolors.grey}!!{bcolors.reset}" if twice_diff_f1   / n_loops > 0.99 else ''
-            col_sign_acc  = f"{bcolors.red}**{bcolors.reset}" if twice_diff_acc  / n_loops < 0.01 else f"{bcolors.red}*{bcolors.reset}" if twice_diff_acc  / n_loops < 0.05 else f"{bcolors.grey}!{bcolors.reset}" if twice_diff_acc  / n_loops > 0.95 else f"{bcolors.grey}!!{bcolors.reset}" if twice_diff_acc  / n_loops > 0.99 else ''
-            col_sign_prec = f"{bcolors.red}**{bcolors.reset}" if twice_diff_prec / n_loops < 0.01 else f"{bcolors.red}*{bcolors.reset}" if twice_diff_prec / n_loops < 0.05 else f"{bcolors.grey}!{bcolors.reset}" if twice_diff_prec / n_loops > 0.95 else f"{bcolors.grey}!!{bcolors.reset}" if twice_diff_prec / n_loops > 0.99 else ''
-            col_sign_rec  = f"{bcolors.red}**{bcolors.reset}" if twice_diff_rec  / n_loops < 0.01 else f"{bcolors.red}*{bcolors.reset}" if twice_diff_rec  / n_loops < 0.05 else f"{bcolors.grey}!{bcolors.reset}" if twice_diff_rec  / n_loops > 0.95 else f"{bcolors.grey}!!{bcolors.reset}" if twice_diff_rec  / n_loops > 0.99 else ''
+            col_sign_f1   = f"{BColor.red}**{BColor.reset}" if twice_diff_f1   / n_loops < 0.01 else f"{BColor.red}*{BColor.reset}" if twice_diff_f1   / n_loops < 0.05 else f"{BColor.grey}!{BColor.reset}" if twice_diff_f1   / n_loops > 0.95 else f"{BColor.grey}!!{BColor.reset}" if twice_diff_f1   / n_loops > 0.99 else ''
+            col_sign_acc  = f"{BColor.red}**{BColor.reset}" if twice_diff_acc  / n_loops < 0.01 else f"{BColor.red}*{BColor.reset}" if twice_diff_acc  / n_loops < 0.05 else f"{BColor.grey}!{BColor.reset}" if twice_diff_acc  / n_loops > 0.95 else f"{BColor.grey}!!{BColor.reset}" if twice_diff_acc  / n_loops > 0.99 else ''
+            col_sign_prec = f"{BColor.red}**{BColor.reset}" if twice_diff_prec / n_loops < 0.01 else f"{BColor.red}*{BColor.reset}" if twice_diff_prec / n_loops < 0.05 else f"{BColor.grey}!{BColor.reset}" if twice_diff_prec / n_loops > 0.95 else f"{BColor.grey}!!{BColor.reset}" if twice_diff_prec / n_loops > 0.99 else ''
+            col_sign_rec  = f"{BColor.red}**{BColor.reset}" if twice_diff_rec  / n_loops < 0.01 else f"{BColor.red}*{BColor.reset}" if twice_diff_rec  / n_loops < 0.05 else f"{BColor.grey}!{BColor.reset}" if twice_diff_rec  / n_loops > 0.95 else f"{BColor.grey}!!{BColor.reset}" if twice_diff_rec  / n_loops > 0.99 else ''
             sign_f1   = "**" if twice_diff_f1   / n_loops < 0.01 else "*" if twice_diff_f1   / n_loops < 0.05 else "!" if twice_diff_f1   / n_loops > 0.95 else "!!" if twice_diff_f1   / n_loops > 0.99 else ''
             sign_acc  = "**" if twice_diff_acc  / n_loops < 0.01 else "*" if twice_diff_acc  / n_loops < 0.05 else "!" if twice_diff_acc  / n_loops > 0.95 else "!!" if twice_diff_acc  / n_loops > 0.99 else ''
             sign_prec = "**" if twice_diff_prec / n_loops < 0.01 else "*" if twice_diff_prec / n_loops < 0.05 else "!" if twice_diff_prec / n_loops > 0.95 else "!!" if twice_diff_prec / n_loops > 0.99 else ''
             sign_rec  = "**" if twice_diff_rec  / n_loops < 0.01 else "*" if twice_diff_rec  / n_loops < 0.05 else "!" if twice_diff_rec  / n_loops > 0.95 else "!!" if twice_diff_rec  / n_loops > 0.99 else ''
-            str_out = f"{'count sample diff f1   is twice tot diff f1':.<60} {twice_diff_f1:<5}/ {n_loops:<8}p < {round((twice_diff_f1 / n_loops), 4):<6} {col_sign_f1}\n" \
-                      f"{'count sample diff acc  is twice tot diff acc':.<60} {twice_diff_acc:<5}/ {n_loops:<8}p < {round((twice_diff_acc / n_loops), 4):<6} {col_sign_acc }\n" \
-                      f"{'count sample diff prec is twice tot diff prec':.<60} {twice_diff_prec:<5}/ {n_loops:<8}p < {round((twice_diff_prec / n_loops), 4):<6} {col_sign_prec}\n" \
-                      f"{'count sample diff rec  is twice tot diff rec ':.<60} {twice_diff_rec:<5}/ {n_loops:<8}p < {round((twice_diff_rec / n_loops), 4):<6} {col_sign_rec }"
+            str_out = f"{'count sample diff f1   is twice tot diff f1':.<50} {twice_diff_f1:<5}/ {n_loops:<8}p < {round((twice_diff_f1 / n_loops), 4):<6} {col_sign_f1}\n" \
+                      f"{'count sample diff prec is twice tot diff prec':.<50} {twice_diff_prec:<5}/ {n_loops:<8}p < {round((twice_diff_prec / n_loops), 4):<6} {col_sign_prec}\n" \
+                      f"{'count sample diff rec  is twice tot diff rec ':.<50} {twice_diff_rec:<5}/ {n_loops:<8}p < {round((twice_diff_rec / n_loops), 4):<6} {col_sign_rec }\n" \
+                      f"{'count sample diff acc  is twice tot diff acc':.<50} {twice_diff_acc:<5}/ {n_loops:<8}p < {round((twice_diff_acc / n_loops), 4):<6} {col_sign_acc }"
             df_tot.s_f1   = ['', sign_f1]
             df_tot.s_acc  = ['', sign_acc]
             df_tot.s_prec = ['', sign_prec]
             df_tot.s_rec  = ['', sign_rec]
             if targetclass is not None:
-                col_sign_tgt_f1   = f"{bcolors.red}**{bcolors.reset}" if twice_diff_tgt_f1   / n_loops < 0.01 else f"{bcolors.red}*{bcolors.reset}" if twice_diff_tgt_f1   / n_loops < 0.05 else f"{bcolors.grey}!{bcolors.reset}" if twice_diff_tgt_f1   / n_loops > 0.95 else f"{bcolors.grey}!!{bcolors.reset}" if twice_diff_tgt_f1   / n_loops > 0.99 else ''
-                col_sign_tgt_prec = f"{bcolors.red}**{bcolors.reset}" if twice_diff_tgt_prec / n_loops < 0.01 else f"{bcolors.red}*{bcolors.reset}" if twice_diff_tgt_prec / n_loops < 0.05 else f"{bcolors.grey}!{bcolors.reset}" if twice_diff_tgt_prec / n_loops > 0.95 else f"{bcolors.grey}!!{bcolors.reset}" if twice_diff_tgt_prec / n_loops > 0.99 else ''
-                col_sign_tgt_rec  = f"{bcolors.red}**{bcolors.reset}" if twice_diff_tgt_rec  / n_loops < 0.01 else f"{bcolors.red}*{bcolors.reset}" if twice_diff_tgt_rec  / n_loops < 0.05 else f"{bcolors.grey}!{bcolors.reset}" if twice_diff_tgt_rec  / n_loops > 0.95 else f"{bcolors.grey}!!{bcolors.reset}" if twice_diff_tgt_rec  / n_loops > 0.99 else ''
+                col_sign_tgt_f1   = f"{BColor.red}**{BColor.reset}" if twice_diff_tgt_f1   / n_loops < 0.01 else f"{BColor.red}*{BColor.reset}" if twice_diff_tgt_f1   / n_loops < 0.05 else f"{BColor.grey}!{BColor.reset}" if twice_diff_tgt_f1   / n_loops > 0.95 else f"{BColor.grey}!!{BColor.reset}" if twice_diff_tgt_f1   / n_loops > 0.99 else ''
+                col_sign_tgt_prec = f"{BColor.red}**{BColor.reset}" if twice_diff_tgt_prec / n_loops < 0.01 else f"{BColor.red}*{BColor.reset}" if twice_diff_tgt_prec / n_loops < 0.05 else f"{BColor.grey}!{BColor.reset}" if twice_diff_tgt_prec / n_loops > 0.95 else f"{BColor.grey}!!{BColor.reset}" if twice_diff_tgt_prec / n_loops > 0.99 else ''
+                col_sign_tgt_rec  = f"{BColor.red}**{BColor.reset}" if twice_diff_tgt_rec  / n_loops < 0.01 else f"{BColor.red}*{BColor.reset}" if twice_diff_tgt_rec  / n_loops < 0.05 else f"{BColor.grey}!{BColor.reset}" if twice_diff_tgt_rec  / n_loops > 0.95 else f"{BColor.grey}!!{BColor.reset}" if twice_diff_tgt_rec  / n_loops > 0.99 else ''
                 sign_tgt_f1   = "**" if twice_diff_tgt_f1   / n_loops < 0.01 else "*" if twice_diff_tgt_f1   / n_loops < 0.05 else "!" if twice_diff_tgt_f1   / n_loops > 0.95 else "!!" if twice_diff_tgt_f1   / n_loops > 0.99 else ''
                 sign_tgt_prec = "**" if twice_diff_tgt_prec / n_loops < 0.01 else "*" if twice_diff_tgt_prec / n_loops < 0.05 else "!" if twice_diff_tgt_prec / n_loops > 0.95 else "!!" if twice_diff_tgt_prec / n_loops > 0.99 else ''
                 sign_tgt_rec  = "**" if twice_diff_tgt_rec  / n_loops < 0.01 else "*" if twice_diff_tgt_rec  / n_loops < 0.05 else "!" if twice_diff_tgt_rec  / n_loops > 0.95 else "!!" if twice_diff_tgt_rec  / n_loops > 0.99 else ''
@@ -478,6 +540,7 @@ class Bootstrap:
             twice_diff_ce  = 0
             twice_diff_sim = 0
             twice_diff_cor = 0
+            diff_tgt_jsd, diff_tgt_ce, twice_diff_tgt_jsd, twice_diff_tgt_ce = None, None, None, None
             if targetclass is not None:
                 diff_tgt_jsd = df_tgt.d_tjsd[-1]
                 diff_tgt_ce  = df_tgt.d_tce[-1]
@@ -496,25 +559,25 @@ class Bootstrap:
                 if targetclass is not None:
                     if df_sample_tgt.d_tjsd[-1] > -2 * diff_tgt_jsd: twice_diff_tgt_jsd += 1 # lower is better
                     if df_sample_tgt.d_tce[-1]  > -2 * diff_tgt_ce:  twice_diff_tgt_ce  += 1 # lower is better
-            col_sign_jsd = f"{bcolors.red}**{bcolors.reset}" if twice_diff_jsd / n_loops < 0.01 else f"{bcolors.red}*{bcolors.reset}" if twice_diff_jsd / n_loops < 0.05 else f"{bcolors.grey}!{bcolors.reset}" if twice_diff_jsd / n_loops > 0.95 else f"{bcolors.grey}!!{bcolors.reset}" if twice_diff_jsd / n_loops > 0.99 else ''
-            col_sign_ce  = f"{bcolors.red}**{bcolors.reset}" if twice_diff_ce  / n_loops < 0.01 else f"{bcolors.red}*{bcolors.reset}" if twice_diff_ce  / n_loops < 0.05 else f"{bcolors.grey}!{bcolors.reset}" if twice_diff_ce  / n_loops > 0.95 else f"{bcolors.grey}!!{bcolors.reset}" if twice_diff_ce  / n_loops > 0.99 else ''
-            col_sign_sim = f"{bcolors.red}**{bcolors.reset}" if twice_diff_sim / n_loops < 0.01 else f"{bcolors.red}*{bcolors.reset}" if twice_diff_sim / n_loops < 0.05 else f"{bcolors.grey}!{bcolors.reset}" if twice_diff_sim / n_loops > 0.95 else f"{bcolors.grey}!!{bcolors.reset}" if twice_diff_sim / n_loops > 0.99 else ''
-            col_sign_cor = f"{bcolors.red}**{bcolors.reset}" if twice_diff_cor / n_loops < 0.01 else f"{bcolors.red}*{bcolors.reset}" if twice_diff_cor / n_loops < 0.05 else f"{bcolors.grey}!{bcolors.reset}" if twice_diff_cor / n_loops > 0.95 else f"{bcolors.grey}!!{bcolors.reset}" if twice_diff_cor / n_loops > 0.99 else ''
+            col_sign_jsd = f"{BColor.red}**{BColor.reset}" if twice_diff_jsd / n_loops < 0.01 else f"{BColor.red}*{BColor.reset}" if twice_diff_jsd / n_loops < 0.05 else f"{BColor.grey}!{BColor.reset}" if twice_diff_jsd / n_loops > 0.95 else f"{BColor.grey}!!{BColor.reset}" if twice_diff_jsd / n_loops > 0.99 else ''
+            col_sign_ce  = f"{BColor.red}**{BColor.reset}" if twice_diff_ce  / n_loops < 0.01 else f"{BColor.red}*{BColor.reset}" if twice_diff_ce  / n_loops < 0.05 else f"{BColor.grey}!{BColor.reset}" if twice_diff_ce  / n_loops > 0.95 else f"{BColor.grey}!!{BColor.reset}" if twice_diff_ce  / n_loops > 0.99 else ''
+            col_sign_sim = f"{BColor.red}**{BColor.reset}" if twice_diff_sim / n_loops < 0.01 else f"{BColor.red}*{BColor.reset}" if twice_diff_sim / n_loops < 0.05 else f"{BColor.grey}!{BColor.reset}" if twice_diff_sim / n_loops > 0.95 else f"{BColor.grey}!!{BColor.reset}" if twice_diff_sim / n_loops > 0.99 else ''
+            col_sign_cor = f"{BColor.red}**{BColor.reset}" if twice_diff_cor / n_loops < 0.01 else f"{BColor.red}*{BColor.reset}" if twice_diff_cor / n_loops < 0.05 else f"{BColor.grey}!{BColor.reset}" if twice_diff_cor / n_loops > 0.95 else f"{BColor.grey}!!{BColor.reset}" if twice_diff_cor / n_loops > 0.99 else ''
             sign_jsd = "**" if twice_diff_jsd / n_loops < 0.01 else "*" if twice_diff_jsd / n_loops < 0.05 else "!" if twice_diff_jsd / n_loops > 0.95 else "!!" if twice_diff_jsd / n_loops > 0.99 else ''
             sign_ce  = "**" if twice_diff_ce  / n_loops < 0.01 else "*" if twice_diff_ce  / n_loops < 0.05 else "!" if twice_diff_ce  / n_loops > 0.95 else "!!" if twice_diff_ce  / n_loops > 0.99 else ''
             sign_sim = "**" if twice_diff_sim / n_loops < 0.01 else "*" if twice_diff_sim / n_loops < 0.05 else "!" if twice_diff_sim / n_loops > 0.95 else "!!" if twice_diff_sim / n_loops > 0.99 else ''
             sign_cor = "**" if twice_diff_cor / n_loops < 0.01 else "*" if twice_diff_cor / n_loops < 0.05 else "!" if twice_diff_cor / n_loops > 0.95 else "!!" if twice_diff_cor / n_loops > 0.99 else ''
-            str_out = f"{'count sample diff jsd is twice tot diff jsd':.<60} {twice_diff_jsd:<5}/ {n_loops:<8}p < {round((twice_diff_jsd / n_loops), 4):<6} {col_sign_jsd}\n" \
-                      f"{'count sample diff ce  is twice tot diff ce':.<60} {twice_diff_ce:<5}/ {n_loops:<8}p < {round((twice_diff_ce / n_loops), 4):<6} {col_sign_ce }\n" \
-                      f"{'count sample diff sim is twice tot diff sim':.<60} {twice_diff_sim:<5}/ {n_loops:<8}p < {round((twice_diff_sim / n_loops), 4):<6} {col_sign_sim}\n" \
-                      f"{'count sample diff cor is twice tot diff cor':.<60} {twice_diff_cor:<5}/ {n_loops:<8}p < {round((twice_diff_cor / n_loops), 4):<6} {col_sign_cor }"
+            str_out = f"{'count sample diff jsd is twice tot diff jsd':.<50} {twice_diff_jsd:<5}/ {n_loops:<8}p < {round((twice_diff_jsd / n_loops), 4):<6} {col_sign_jsd}\n" \
+                      f"{'count sample diff ce  is twice tot diff ce':.<50} {twice_diff_ce:<5}/ {n_loops:<8}p < {round((twice_diff_ce / n_loops), 4):<6} {col_sign_ce }\n" \
+                      f"{'count sample diff sim is twice tot diff sim':.<50} {twice_diff_sim:<5}/ {n_loops:<8}p < {round((twice_diff_sim / n_loops), 4):<6} {col_sign_sim}\n" \
+                      f"{'count sample diff cor is twice tot diff cor':.<50} {twice_diff_cor:<5}/ {n_loops:<8}p < {round((twice_diff_cor / n_loops), 4):<6} {col_sign_cor }"
             df_tot.s_jsd = ['', sign_jsd]
             df_tot.s_ce  = ['', sign_ce]
             df_tot.s_sim = ['', sign_sim]
             df_tot.s_cor = ['', sign_cor]
             if targetclass is not None:
-                col_sign_tgt_jsd = f"{bcolors.red}**{bcolors.reset}" if twice_diff_tgt_jsd / n_loops < 0.01 else f"{bcolors.red}*{bcolors.reset}" if twice_diff_tgt_jsd / n_loops < 0.05 else f"{bcolors.grey}!{bcolors.reset}" if twice_diff_tgt_jsd / n_loops > 0.95 else f"{bcolors.grey}!!{bcolors.reset}" if twice_diff_tgt_jsd / n_loops > 0.99 else ''
-                col_sign_tgt_ce  = f"{bcolors.red}**{bcolors.reset}" if twice_diff_tgt_ce  / n_loops < 0.01 else f"{bcolors.red}*{bcolors.reset}" if twice_diff_tgt_ce  / n_loops < 0.05 else f"{bcolors.grey}!{bcolors.reset}" if twice_diff_tgt_ce  / n_loops > 0.95 else f"{bcolors.grey}!!{bcolors.reset}" if twice_diff_tgt_ce  / n_loops > 0.99 else ''
+                col_sign_tgt_jsd = f"{BColor.red}**{BColor.reset}" if twice_diff_tgt_jsd / n_loops < 0.01 else f"{BColor.red}*{BColor.reset}" if twice_diff_tgt_jsd / n_loops < 0.05 else f"{BColor.grey}!{BColor.reset}" if twice_diff_tgt_jsd / n_loops > 0.95 else f"{BColor.grey}!!{BColor.reset}" if twice_diff_tgt_jsd / n_loops > 0.99 else ''
+                col_sign_tgt_ce  = f"{BColor.red}**{BColor.reset}" if twice_diff_tgt_ce  / n_loops < 0.01 else f"{BColor.red}*{BColor.reset}" if twice_diff_tgt_ce  / n_loops < 0.05 else f"{BColor.grey}!{BColor.reset}" if twice_diff_tgt_ce  / n_loops > 0.95 else f"{BColor.grey}!!{BColor.reset}" if twice_diff_tgt_ce  / n_loops > 0.99 else ''
                 sign_tgt_jsd = "**" if twice_diff_tgt_jsd / n_loops < 0.01 else "*" if twice_diff_tgt_jsd / n_loops < 0.05 else "!" if twice_diff_tgt_jsd / n_loops > 0.95 else "!!" if twice_diff_tgt_jsd / n_loops > 0.99 else ''
                 sign_tgt_ce  = "**" if twice_diff_tgt_ce  / n_loops < 0.01 else "*" if twice_diff_tgt_ce  / n_loops < 0.05 else "!" if twice_diff_tgt_ce  / n_loops > 0.95 else "!!" if twice_diff_tgt_ce  / n_loops > 0.99 else ''
                 str_out += f"\ntarget {targetclass} {'count sample diff jsd is twice tot diff jsd':.<50} {twice_diff_tgt_jsd:<5}/ {n_loops:<8}p < {round((twice_diff_tgt_jsd / n_loops), 4):<6} {col_sign_tgt_jsd}\n" \
@@ -529,9 +592,14 @@ class Bootstrap:
 
     def run(self, n_loops=1000, sample_size=.1, targetclass=None, verbose=False):
         """
-        :param data:
-                defaultdict(lambda: {'exp_idxs': list(), 'preds': list(), 'targs': list(), 'idxs': list(), 'epochs': list(),
-                                     'h1': defaultdict(lambda: {'exp_idxs': list(), 'preds': list(), 'targs': list(), 'idxs': list(), 'epochs': list()})}
+        :param n_loops: type: int; default: 1000. number of iterations for bootstrap sampling.
+        :param sample_size: type: float; default: .1. Percent sample size wrt the test set, f
+        or bootstrap sampling.
+        :param targetclass: type: int; default: None. Index of a target class. If given, performance and
+        significance levels are computed for that class.
+        :param verbose: type: bool; default: False. If true, it prints performance information.
+        :return: two pandas DataFrame: the first contain the overall performance,
+        the second the target class performance (empty if not requested)
         """
         startime = start()
 
@@ -550,14 +618,14 @@ class Bootstrap:
                     if targetclass is not None:
                         h0_f1tgt_all.append(precision_recall_fscore_support(targs, preds)[2][targetclass])
                     if verbose:
-                        print(f"{exp_idx:<60} F1 {f1}")
+                        print(f"{exp_idx:<60} F1 {f1:.4f}")
                 else:
                     jsd = jensenshannon(targs, preds, axis=1).mean()
                     h0_jsd_all.append(jsd)
                     if targetclass is not None:
                         h0_jsdtgt_all.append(jensenshannon(targs[:, targetclass].reshape(1, -1), preds[:, targetclass].reshape(1, -1), axis=1).mean())
                     if verbose:
-                        print(f"{exp_idx:<60} jsd {jsd}")
+                        print(f"{exp_idx:<60} jsd {jsd:.4f}")
 
             for h1_cond in self.data[h0_cond]['h1']:
                 print(f"{'#'*80}\n{h0_cond}   vs   {h1_cond}")
@@ -576,16 +644,16 @@ class Bootstrap:
                         if targetclass is not None:
                             h1_f1tgt_all.append(precision_recall_fscore_support(targs, preds)[2][targetclass])
                         if verbose:
-                            print(f"{exp_idx:<60} F1 {f1}")
+                            print(f"{exp_idx:<60} F1 {f1:.4f}")
                     else:
                         jsd = jensenshannon(targs, preds, axis=1).mean()
                         h1_jsd_all.append(jsd)
                         if targetclass is not None:
                             h1_jsdtgt_all.append(jensenshannon(targs[:, targetclass].reshape(1, -1), preds[:, targetclass].reshape(1, -1), axis=1).mean())
                         if verbose:
-                            print(f"{exp_idx:<60} jsd {jsd}")
-                assert (h0_targs_all == h1_targs_all).all(), 'h0 and h1 targets differ'
-                assert (h0_idxs_all == h1_idxs_all).all(), 'h0 and h1 idxs differ'
+                            print(f"{exp_idx:<60} jsd {jsd:.4f}")
+                assert np.array(h0_targs_all == h1_targs_all).all(), 'h0 and h1 targets differ'
+                assert np.array(h0_idxs_all == h1_idxs_all).all(), 'h0 and h1 idxs differ'
                 targs_all = h0_targs_all
                 
                 df_tot_cond, df_tgt_cond = self.test(targs_all, h0_preds_all, h1_preds_all, h0_name=h0_cond, h1_name=h1_cond, n_loops=n_loops, sample_size=sample_size, targetclass=targetclass, verbose=True)
@@ -637,6 +705,3 @@ class Bootstrap:
             df_tot.to_csv(f"{self.dirout}results.tsv")
         end(startime)
         return df_tot, df_tgt
-
-
-
